// utilities.h
// Gregory Rosenblatt

#ifndef Utilities_H_
#define Utilities_H_

#include <string>
#include <vector>
#include <stdio.h>
#include <assert.h>

namespace MetaLang
{
    // encapsulates a byte array
    class Buffer
    {
    public:
        Buffer() : _size(0), _data(0)   {}
        Buffer(const char* data, size_t size)
            : _size(size), _data(new char[_size])
        { memcpy(_data, data, _size); }
        Buffer(const Buffer& buf) : _size(buf._size), _data(new char[_size])
        { memcpy(_data, buf._data, _size); }
        ~Buffer()   { delete[] _data; }
        void operator=(const Buffer& buf)
        {
            if (&buf == this)
                return;
            delete[] _data;
            _size = buf._size;
            _data = new char[_size];
            memcpy(_data, buf._data, _size);
        }
        const char* Data() const    { return _data; }
    private:
        size_t  _size;
        char*   _data;
    };
    
    // stack for pushing and popping of data groups by scope
    template <class T>
    class ScopeStack
    {
    public:
        ScopeStack() : _stackBase(0)    {}
        ScopeStack(size_t reserveSize) : _stackBase(0)
            { _stack.reserve(reserveSize); }
        // scope interface
        void SetInitialScope(size_t scopeSize)
        {
            assert(_stack.empty() && _indices.empty()); // initial scope should only be set once
            _nextBase = _stackBase+scopeSize;           // preemptively set the next offset
            _stack.resize(_nextBase);                   // accommodate initial scope data
            _indices.push_back(_stackBase);             // register the base offset
        }        
        void PushScope(size_t scopeSize)
        {
            _stackBase = _nextBase;                 // get new scope offset
            _indices.push_back(_stackBase);         // register this offset
            _nextBase = _stackBase+scopeSize;       // store the next index
            _stack.resize(_nextBase);               // accommodate new scope data
        }
        void PopScope()
        {
            assert(!_indices.empty());      // don't try to pop a non-existent scope
            _stack.resize(_stackBase);      // shrink back to the end of the former scope
            _nextBase = _stackBase;         // store what the next index would be
            _indices.pop_back();            // take that index off the index stack
            _stackBase = _indices.back();   // grab the base index for the former scope
        }
        // element interface
        void Push(const T& element) { _stack.push_back(element); }
        void Pop()  { assert(!_stack.empty()); _stack.pop_back(); }
        // accessors
        const T& Top() const    { assert(!_stack.empty()); return _stack.back(); }
        T& Top()    { assert(!_stack.empty()); return _stack.back(); }
        const T& Get(size_t index) const    { assert((_stackBase+index)<_stack.size()); return _stack[_stackBase+index]; }
        T& Get(size_t index)    { assert((_stackBase+index)<_stack.size()); return _stack[_stackBase+index]; }
        const T& GetAtDepth(size_t index, size_t depth) const   { assert((GetScopeIndex(depth)+index)<_stack.size()); return _stack[GetScopeIndex(depth)+index]; }
        T& GetAtDepth(size_t index, size_t depth)   { assert((GetScopeIndex(depth)+index)<_stack.size()); return _stack[GetScopeIndex(depth)+index]; }
        const T& operator[](size_t index) const { return Get(index); }
        T& operator[](size_t index) { return Get(index); }
        // properties
        size_t Size() const { return _stack.size(); }
        size_t ScopeDepth() const   { return _indices.size(); }
        size_t GetScopeIndex(size_t depth) const  { assert(depth<_indices.size()); return _indices[depth]; }
    private:
        std::vector<T>      _stack;
        std::vector<size_t> _indices;
        size_t              _stackBase;
        size_t              _nextBase;
    };
    
    template <class T, typename ReturnType, typename ArgType>
    class Callback
    {
    public:
        typedef std::vector<ArgType>    ArgVector;
        typedef ReturnType (T::*Func)(const ArgVector&);
        explicit Callback(Func f) : _f(f)   {}
        ReturnType operator()(T& t, const ArgVector& argv) const   { return (t.*_f)(argv); }
    private:
        Func    _f;
    };
    
    template <class T, typename ReturnType, typename ArgType>
    Callback<T, ReturnType, ArgType> make_callback(ReturnType (T::*func)(const std::vector<ArgType>&))
    {
        return Callback<T, ReturnType, ArgType>(func);
    }
}

#endif

